import * as assert from "node:assert/strict";
import { after, before, suite, test } from "node:test";
import * as http from "node:http";
import { broadcastMessage, createEventServer } from "./event-server.js";
import { once } from "node:events";
import type { AddressInfo } from "node:net";

suite("function createEventServer", () => {
	let server: ReturnType<typeof createEventServer>;
	let baseUrl: URL;

	before(async () => {
		server = createEventServer();
		server.httpServer.listen(0, "localhost");
		await once(server.httpServer, "listening");
		baseUrl = new URL(`http://localhost:${(server.httpServer.address() as AddressInfo).port}`)
	});
	after(() => {
		server.httpServer.close();
	})

	async function streamRequest(method: string, ref: string) {
		const url = new URL(ref, baseUrl);
		const request = http.request(url, {
			method,
			timeout: 1000,
		});
		request.end();
		const response: http.IncomingMessage = (await once(request, "response"))[0];
		return { request, response };
	}

	async function makeRequest(method: string, ref: string) {
		type WithBody = Awaited<ReturnType<typeof streamRequest>> & { response: { body: Buffer } };
		const { request, response } = await streamRequest(method, ref) as WithBody;
		response.body = Buffer.concat(await response.toArray());
		return { request, response };
	}

	test("GET /events", async () => {
		const url = new URL("/events", baseUrl);
		const { response } = await streamRequest("GET", "/events");
		assert.equal(response.statusCode, 200);
		assert.equal(response.headers["content-type"], "text/event-stream");
		broadcastMessage(server, "test");
		broadcastMessage(server, "exit");
		for (const stream of server.streams) {
			stream.end();
		}
		const data = Buffer.concat(await response.toArray()).toString();
		assert.equal(data, "event: test\ndata\n\nevent: exit\ndata\n\n");
	});

	async function badRequestTest(method: string) {
		const { response } = await makeRequest(method, "/bad-request");
		const content = Buffer.from("400 Bad Request");
		assert.equal(response.statusCode, 400);
		assert.equal(response.headers["content-type"], "text/plain");
		assert.equal(Number.parseInt(response.headers["content-length"]!), content.length);
		assert.deepEqual(response.body, method === "HEAD" ? Buffer.alloc(0) : content);
	}

	test("GET /bad-request", async () => {
		await badRequestTest("GET");
	});

	test("HEAD /bad-request", async () => {
		await badRequestTest("HEAD");
	});

	test("POST /bad-request", async () => {
		await badRequestTest("POST");
	});
});
